#include <mpi.h>
#include <fstream>
#include <iterator>
#include <vector>
#include <string>
#include <iostream>
#include <cstdio>
#include <stdexcept>
#include <cstdint>
#include <limits>
#include <algorithm>
#include <numeric>
#include <map>
#include "utils.hpp"
#include "io.hpp"

#include "util/prefix_sum_tools.hpp"
#include "oned/nicol_plus.hpp"

size_t block_decompose(const size_t n, const int p, const int rank)
{
    return n / p + (((size_t)rank < n % p) ? 1 : 0);
}
std::vector<uint64_t> read_custom_decomposition(const char* mapping_file_path, const int nProc) {
    FILE *f = fopen(mapping_file_path, "r");
    std::vector <uint64_t> procStarts;
    if (f == NULL){
        throw std::runtime_error(std::string("Couldn't open file ") + mapping_file_path);
    }
    procStarts.resize(nProc);
    int i = 0;
    uint64_t tmp = -1;
    while (fscanf (f, " %llu", &tmp) != EOF) {
        if (i == nProc) {
            std::cout << "there are more than required number of numbers! nProc = " << nProc << std::endl;
        }
        procStarts[i++] = tmp;
    }
    return procStarts;
}

// template <class uint64_t>
void write_custom_decomposition(std::vector<uint64_t>& all_start_indices, std::string fn) {
    FILE *f = fopen(fn.c_str(), "w");
    for (auto i: all_start_indices)
        fprintf(f, "%llu ", i);
    fprintf(f, "\n");
    fclose(f);
}

std::vector<uint64_t> getDecomposition(std::vector<uint64_t>& hist_main,
                        const char* mapping_file_path, int mode, int source, const int rank, const int nProc, sizeType alpha) {
    // printf("alpha in getDecomposition %lli\n", alpha);
    std::vector<uint64_t> all_local_sizes;
    if (mapping_file_path != NULL) {
        if (rank == source) {
            all_local_sizes = read_custom_decomposition(mapping_file_path, nProc);
        }
        else { // if (rank != source) {
            all_local_sizes.resize(nProc);
        }
        MPI_Bcast(all_local_sizes.data(), nProc, MPI_INT, source, MPI_COMM_WORLD);
    }
    else {
        all_local_sizes.resize(nProc);

        const int highestK = hist_main.size() - 2;

        bool lastFits = 0;
        if (false) {}
        else {
            std::vector<uint64_t> hist(hist_main.begin(), hist_main.end());
            std::vector<sizeType> lo(nProc);
            sizeType cnt = 0;
            sizeType totLoad = 0;
            sizeType avgLoad = -1;

            sizeType loadOfLastCore = hist[highestK] * (highestK+alpha);

            int r;
            for (r = 0; r <= highestK; ++r) {
                totLoad += (r+alpha) * hist[r];
            }

            avgLoad = totLoad / nProc + (totLoad % nProc == 0 ? 0 : 1);

            if (loadOfLastCore > avgLoad) {
                // Decrease the load for the last core from total. we process the separately at the end.
                #ifdef PRINT
                std::cout << "last kcore does not fit in one proc in edge-block decompose hist = " << hist[highestK] << " , loadOfLastCore = " << loadOfLastCore << " , avgLoad = " << avgLoad << std::endl;
                #endif
                totLoad -= ((highestK+alpha) * (highestK+1));
                hist[highestK] -= (highestK+1);
            }
            else { // if (loadOfLastCore <= avgLoad) {
                #ifdef PRINT
                std::cout << "last kcore fits in one proc in edge-block decompose hist = " << hist[highestK] << " , loadOfLastCore = " << loadOfLastCore << " , avgLoad = " << avgLoad << std::endl;
                #endif
                lastFits = 1;
            }

            avgLoad = totLoad / nProc + (totLoad % nProc == 0 ? 0 : 1);

            #ifdef PRINT
            std::cout << "totLoad " << totLoad << " avgLoad " << avgLoad << std::endl;
            #endif
            r = 0;
            sizeType remLoad = avgLoad;
            int curK = 0;
            sizeType diff = 0;
            while (totLoad > 0) {

                if ((sizeType) (hist[curK] * (curK+alpha)) <= remLoad) {

                    if (curK == highestK && r == DEF_lastProc - 1){
                        remLoad = avgLoad;
                        ++r;
                    }
                    remLoad -= hist[curK] * (curK+alpha);
                    totLoad -= hist[curK] * (curK+alpha);
                    lo[r] += hist[curK] * (curK+alpha);
                    all_local_sizes[r] += hist[curK];
                    hist[curK] = 0;
                    #ifdef PRINT
                    std::cout << "Proc " << r << " loaded k " << curK << " fully. remload " << remLoad << " localsize " << all_local_sizes[r] << std::endl;
                    #endif
                    ++curK;
                    if (curK > highestK) {
                        break;
                    }
                }
                else {
                    if (lastFits && r < DEF_lastProc && curK == highestK) {
                        // Last one fits the last process, we are not at the last process,
                        // we are at the last K-core,
                        // we are trying to put some part of it to one of the other processes.
                        #ifdef PRINT
                        std::cout << "proc " << r << " can load some of last core but does not because it can fit in last proc!" << std::endl;
                        #endif
                        remLoad = avgLoad;
                        ++r;
                        continue;
                    }
                    cnt = (remLoad / (curK+alpha));
                    diff = remLoad - cnt * (curK+alpha);
                    if (diff > 0) // not equal
                        ++cnt;

                    remLoad -= cnt * (curK+alpha);
                    totLoad -= cnt * (curK+alpha);
                    lo[r] += cnt * (curK+alpha);
                    all_local_sizes[r] += cnt;
                    hist[curK] -= cnt;
                    #ifdef PRINT
                    std::cout << "remaining hist[" << curK << "] is bigger than avg load.. cnt = " << cnt << " proc " << r << " remHist[curK] = " << hist[curK] << std::endl;
                    #endif

                    if (totLoad == 0) {
                        break;
                    }
                    else if (totLoad < 0) {
                        std::cout << "distributed more load than in the job! " << totLoad << std::endl;
                        std::cout << std::accumulate(all_local_sizes.begin(), all_local_sizes.end(), uint64_t()) << " r " << r << " curK " << curK << std::endl;

                        MPI_Abort(MPI_COMM_WORLD, 1);
                    }
                    remLoad = avgLoad;

                    #ifdef PRINT
                    std::cout << "Proc " << r << " is done. nnodes " << all_local_sizes[r] << std::endl;
                    #endif

                    ++r;
                    if (r >= nProc){
                        std::cout << "could not divide!" << std::endl;
                        std::cout <<"total distributed " << std::accumulate(all_local_sizes.begin(), all_local_sizes.end(), uint64_t()) << std::endl;
                        MPI_Abort(MPI_COMM_WORLD, 1);
                    }
                }
            }
            if (rank == DEF_lastProc) {
                std::cout <<"Greedy MAX " << *std::max_element(lo.begin(),lo.end()) << std::endl;
                std::cout << all_local_sizes << std::endl;
            }
            if ( !lastFits ) {
                all_local_sizes[DEF_lastProc] += (highestK+1); // to the last processor.
            }

        }
    }
    return all_local_sizes;
}

std::vector<int> generate_local_vector(const uint64_t local_size, std::vector<uint64_t>& all_start_indices, std::vector<uint64_t>& hist_starti, const int rank){
    // hist_starti is prefix sum
    // for each one, I know where to start and how much to take.
    std::vector<int> local_elements;
    // printf("local size %d\n", local_size);
    local_elements.reserve(local_size);
    uint64_t rem = local_size;
    int start_k = std::upper_bound(hist_starti.begin(), hist_starti.end(), all_start_indices[rank]) - hist_starti.begin() - 1;
    uint64_t prevEnd = all_start_indices[rank]; // the index previous rank proc ends at. (is prevEnd-1)
    // printf("rank %d starts with k %d\n", rank, start_k);
    while (rem > 0) {
        // printf("rank %d current k %d max (..) %d rem %d\n", rank, start_k, std::max(hist_starti[start_k], prevEnd), rem);
        if (hist_starti[start_k+1] - std::max(hist_starti[start_k], prevEnd) <= rem) {
            // printf("rank %d filling with k %d\n", rank, start_k);
            local_elements.insert(local_elements.end(), // from this index
                hist_starti[start_k+1] - std::max(hist_starti[start_k], prevEnd), // this many
                start_k); // this value
            rem -= hist_starti[start_k+1] - std::max(hist_starti[start_k], prevEnd);
            ++start_k;
        }
        else {
            // printf("rank %d finished with k %d\n", rank, start_k);
            local_elements.insert(local_elements.end(), rem, start_k);
            rem = 0;
        }
    }
    return local_elements;
}

std::vector< std::pair <uint64_t, std::vector <std::pair <int, uint64_t> > > >
            generateKToProcMapping(std::vector<uint64_t>& all_local_sizes, std::vector<uint64_t>& all_start_indices, std::vector<uint64_t>& hist_starti, const int rank){

    // NOTE: 99999 is pid for where the k values end. chosen 99999 to make sure we cannot mix it up with actual process ids.

    std::vector< std::pair <uint64_t, std::vector <std::pair <int, uint64_t> > > > local_info; // for each k, vector : <proc id, start index of proc for the k>
    uint32_t start_k = std::upper_bound(hist_starti.begin(), hist_starti.end(), all_start_indices[rank]) - hist_starti.begin() - 1;
    uint32_t end_k = std::upper_bound(hist_starti.begin(), hist_starti.end(), all_start_indices[rank] + all_local_sizes[rank] - 1) - hist_starti.begin() - 1;

    int kit = 0;
    local_info.push_back ( std::make_pair (start_k, std::vector <std::pair <int, uint64_t> > () ) );
    size_t locationOfStart = upper_bound(all_start_indices.begin(), all_start_indices.end(), hist_starti[start_k]) - all_start_indices.begin() - 1;
    uint64_t remInHist = hist_starti[start_k + 1] - hist_starti[start_k]; // count of this k
    int startForNextProc = 0;
    while (locationOfStart < all_local_sizes.size() && all_start_indices[locationOfStart] < hist_starti[start_k + 1]) { // while current proc can take elements from current k
        if (locationOfStart >= (size_t) rank)
            local_info[kit].second.push_back ( std::make_pair (locationOfStart, startForNextProc) );
        int assignedAmount = std::min(
                (
                    all_local_sizes[locationOfStart] + all_start_indices[locationOfStart] // my end index
                    - std::max(all_start_indices[locationOfStart], hist_starti[start_k]) //minus my start index or start of k
                ) // --> resulting: my capacity for this k
                ,remInHist);
        remInHist -= assignedAmount;
        startForNextProc += assignedAmount;
        ++locationOfStart;
    }

    local_info[kit].second.push_back (std::make_pair (99999, hist_starti[start_k +1] - hist_starti[start_k]));
    if (start_k != end_k) {
        ++start_k;
        ++kit;
        while (start_k < end_k) { // k's in between
            local_info.push_back ( std::make_pair (start_k, std::vector <std::pair <int, uint64_t> > () ) );
            local_info[kit].second.push_back ( std::make_pair (rank, 0) );
            local_info[kit].second.push_back ( std::make_pair (99999, hist_starti[start_k+1] - hist_starti[start_k]) );
            ++kit;
            ++start_k;
        }
        // last k in proc
        local_info.push_back ( std::make_pair (end_k, std::vector <std::pair <int, uint64_t> > () ) );
        locationOfStart = rank; // end_k starts at us
        remInHist = hist_starti[end_k + 1] - hist_starti[end_k]; // count of this k
        startForNextProc = 0;
        while (locationOfStart < all_local_sizes.size() && all_start_indices[locationOfStart] < hist_starti[end_k + 1]) {
        // while current proc can take elements from current k
            local_info[kit].second.push_back ( std::make_pair (locationOfStart, startForNextProc) );
            int assignedAmount = std::min(
                (
                    all_local_sizes[locationOfStart] + all_start_indices[locationOfStart] // my end index
                    - std::max(all_start_indices[locationOfStart], hist_starti[end_k]) //minus my start index or start of k
                ) // --> resulting: my capacity for this k
                ,
                remInHist); // or the rest of this k
            remInHist -= assignedAmount;
            startForNextProc += assignedAmount;
            ++locationOfStart;
        }
        local_info[kit].second.push_back (std::make_pair (99999, hist_starti[end_k +1] - hist_starti[end_k]));
    }

    return local_info;
}
// A modified version: Last processor gets all the n-highestK-1 nodes.
std::vector<int> scatter_vector_custom_decomp(std::vector<int>& global_vec, //std::vector<int>& hist_starti,
                        std::vector<int>& all_local_sizes, int source, MPI_Comm comm) {
    // get MPI Communicator properties
    int rank, nProc;
    MPI_Comm_size(comm, &nProc);
    MPI_Comm_rank(comm, &rank);

    // init result
    std::vector<int> local_elements(all_local_sizes[rank]);

    if (rank == source)
    {
        /* I am the root process */
        // scatter-v the actual data
        std::vector<int> counts(nProc);
        std::vector<int> displs(nProc);
        displs[0] = 0;
        for (int i = 0; i < nProc; ++i)
        {
            counts[i] = all_local_sizes[i];
            if (i > 0)
                displs[i] = displs[i-1] + counts[i-1];
        }
        MPI_Scatterv(&global_vec[0], &counts[0], &displs[0], MPI_INT,
                     &local_elements[0], all_local_sizes[rank], MPI_INT, source, comm);
    }
    else
    {
        /* I am NOT the root process */
        // actually receive all the data
        MPI_Scatterv(NULL, NULL, NULL, MPI_INT,
                     &local_elements[0], all_local_sizes[rank], MPI_INT, source, comm);
    }
    // return local array
    return local_elements;

}

template <typename index_t = int>
std::vector<index_t> get_displacements(const std::vector<index_t>& counts)
{
    // copy and do an exclusive prefix sum
    std::vector<index_t> result(counts);
    // set the total sum to zero
    index_t sum = 0;
    index_t tmp;

    // calculate the exclusive prefix sum
    typename std::vector<index_t>::iterator begin = result.begin();
    while (begin != result.end())
    {
        tmp = sum;
        // assert that the sum will still fit into the index type (MPI default:
        // int)
        sum += *begin;
        *begin = tmp;
        ++begin;
    }
    return result;
}

std::vector<char> gatherNames(const char* data, int size, int root, MPI_Comm& comm) {
    std::vector<int> recv_sizes;
    std::vector<char> result;
    int rank, nProc;
    MPI_Comm_rank(comm, &rank);
    MPI_Comm_size(comm, &nProc);
    if (rank == root)
        recv_sizes.resize(size * nProc);
    MPI_Gather(&size, 1, MPI_INT, &recv_sizes[0], 1, MPI_INT, root, comm);

    size_t total_size = std::accumulate(recv_sizes.begin(), recv_sizes.end(), 0);
    MPI_Bcast(&total_size, 1, MPI_UNSIGNED_LONG, root, comm);
    if (rank == root) {
        result.resize(total_size);
        std::vector<int> displs = get_displacements(recv_sizes);
        MPI_Gatherv(data, size, MPI_CHAR, &result[0], &recv_sizes[0],
                    &displs[0], MPI_CHAR, root, comm);
    } else {
        MPI_Gatherv(data, size, MPI_CHAR, NULL, NULL,
                    NULL, MPI_CHAR, root, comm);
    }
    return result;
}

void print_node_distribution(MPI_Comm comm)
{
    // get MPI parameters
    int rank;
    MPI_Comm_rank(comm, &rank);
    int p;
    MPI_Comm_size(comm, &p);

    // get local processor name
    char p_name[MPI_MAX_PROCESSOR_NAME+1];
    int p_len;
    MPI_Get_processor_name(p_name, &p_len);
    p_name[p_len] = '\0'; // make string NULL-terminated (if not yet so)

    // gather all processor names to master
    std::vector<char> all_names_raw = gatherNames(p_name,p_len+1, 0,comm);

    if (rank == DEF_firstProc) {
        std::vector<std::string> all_names(p);
        // disect the names into a set of strings
        std::vector<char>::iterator str_start = all_names_raw.begin();
        int i = 0;
        for (auto it = all_names_raw.begin(); it != all_names_raw.end(); ++it) {
            if (*it == '\0') {
                all_names[i++] = std::string(str_start, it);
                str_start = it+1;
            }
        }
        assert(i == p);
        std::map<std::string, std::vector<int> > procs_per_node;
        for (i = 0; i < p; ++i) {
            procs_per_node[all_names[i]].push_back(i);
        }
        // create array instead of map, then we can sort by first rank
        std::vector<std::pair<std::string, std::vector<int> > > proc_distr(procs_per_node.size());
        i = 0;
        for (auto it = procs_per_node.begin(); it != procs_per_node.end(); ++it) {
            // sort procs on each node
            std::sort(it->second.begin(), it->second.end());
            // put into the vector
            proc_distr[i++] = std::make_pair(it->first, it->second);
        }
        // sort the vector of node names by first rank
        std::sort(proc_distr.begin(), proc_distr.end(),
                  [](const std::pair<std::string, std::vector<int> >& x,
                     const std::pair<std::string, std::vector<int> >& y)
                  { return x.second.front() < y.second.front();});

        // print out the rank distribution
        std::cerr << "== Node distribution == " << std::endl;
        std::cerr << "== p=" << p << " processes on " << proc_distr.size() << " nodes ==" << std::endl;
        for (auto it = proc_distr.begin(); it != proc_distr.end(); ++it)
        {
            std::cerr << "--  Node: '" << it->first << "' (" << it->second.size() << "/" << p << ")" << std::endl;
            std::cerr << "        Ranks: ";
            for (auto rank_it = it->second.begin(); rank_it != it->second.end(); ++rank_it)
            {
                std::cerr << *rank_it;
                if (rank_it+1 != it->second.end())
                    std::cerr << ", ";
            }
            std::cerr << std::endl;
        }
    }
    return;
}

