/**
 * @file    io.h
 * @author  myozka
 * @brief   Implements common IO and input generation functions.
 *
 */
#ifndef IO_HPP
#define IO_HPP
#include <mpi.h>
#include <fstream>
#include <iterator>
#include <vector>
#include <string>
#include <iostream>
#include <cstdio>
#include <cstdint>
#include <stdexcept>
#include <limits>
#include <algorithm>
#include <numeric>
#include "utils.hpp"

#define NODEBAL 0
#define EDGEBAL 1
#define RELAXEDBAL 2
#define OPTIMALBAL 5
#define GUARANTEEDDIS 999

typedef struct _times{
    double all;
    double generator;
    double communications;
    double fix;
    double phase1;
    double phase2;
    double fixCSR;
    double subset;
    double fullCSR;
    int countOfCascade;
} timeType;

size_t block_decompose(const size_t n, const int p, const int rank);

std::vector<uint64_t> read_custom_decomposition(const char* mapping_file_path, const int nProc);

// template <class uint64_t>
void write_custom_decomposition(std::vector<uint64_t>& all_start_indices, std::string fn);

// template <class uint64_t>
std::vector<uint64_t> getDecomposition(std::vector<uint64_t>& hist_main,
                        const char* mapping_file_path, int mode, int source, const int rank, const int nProc, sizeType alpha);

// template <class uint64_t>
std::vector<int> generate_local_vector(const uint64_t local_size, std::vector<uint64_t>& all_start_indices, std::vector<uint64_t>& hist_starti, const int rank);

// template <class uint64_t>
std::vector< std::pair <uint64_t, std::vector <std::pair <int, uint64_t> > > >
            generateKToProcMapping(std::vector<uint64_t>& all_local_sizes, std::vector<uint64_t>& all_start_indices, std::vector<uint64_t>& hist_starti, const int rank);

// A modified version: Last processor gets all the n-highestK-1 nodes.
std::vector<int> scatter_vector_custom_decomp(std::vector<int>& global_vec, //std::vector<int>& hist_starti,
                        std::vector<int>& all_start_indices, int source, MPI_Comm comm);

void print_node_distribution(MPI_Comm comm = MPI_COMM_WORLD);
#endif // IO_HPP
