#include "mpi_comm.h"

// Returns globalSent
sizeType pointToPointComm_SingleComm (const int rank, const uint64_t start_index, const int countOfCascade, const uint64_t countOfSmallestKLocal,
                            MPI_Comm& smallestKComm,
                            std::vector<int>& procReceivesFrom, std::vector<int>& procSendsTo,
                            std::vector<int>& recvCnt, std::vector<int>& sendCnt,
                            std::vector<std::unique_ptr<uint64_t[]>>& p2precvData, std::vector<std::unique_ptr<uint64_t[]>>& p2psendData,
                            std::priority_queue < uint64_t, std::vector<uint64_t>, std::greater<uint64_t> >& cascadingLocalUpdatesQueue,
                            flat_hash_map <uint64_t, int>& valTable_cascadingLocalUpdatesQueue,
                            std::vector< flat_hash_map< uint64_t, int > >& valTable_cascadingOutOfProcUpd,
                            std::vector<int>& newp) {
    // printf("rank %d START OF pointToPointComm_SingleComm rep %d\n", rank, countOfCascade);
    // First, Receive procedures
    sizeType totalSend = 0;
    sizeType globalSent = 0;
    int recvReqCount = 0, sendReqCount = 0;
    MPI_Request *recvReq = nullptr;
    MPI_Request *sendReq = nullptr;
    MPI_Status status;
    if (procReceivesFrom.size())
        recvReq = new MPI_Request[procReceivesFrom.size()];
    if (procSendsTo.size())
        sendReq = new MPI_Request[procSendsTo.size()];

    size_t indexInArr = 0;
    size_t countDown = procReceivesFrom.size();
    for (int fromProc: procReceivesFrom) {
        indexInArr = fromProc - procReceivesFrom[0];
        MPI_Irecv(&p2precvData[indexInArr][0], 2 * countOfSmallestKLocal,
                    MPI_UINT64_T, fromProc, DATA_TAG+fromProc, MPI_COMM_WORLD, &recvReq[recvReqCount++]);
    }
    // printf ("rank %d posted MPI_Irecvs\n", rank);

    if (procSendsTo.size()) {
        for (int toProc: procSendsTo) {
            indexInArr = (size_t) (toProc - procSendsTo[0]);
            sendCnt[indexInArr] = 2 * valTable_cascadingOutOfProcUpd[toProc].size();
            size_t itSend = 0;
            p2psendData[indexInArr] = std::unique_ptr<uint64_t[]> (new uint64_t[sendCnt[indexInArr]+1]); // +1 for empty send
            totalSend += sendCnt[indexInArr];
            for(auto& mm: valTable_cascadingOutOfProcUpd[toProc]) {
                p2psendData[indexInArr][itSend++] = mm.first; // global node id
                p2psendData[indexInArr][itSend++] = mm.second; // value
            }
            MPI_Isend(&p2psendData[indexInArr][0], sendCnt[indexInArr],
                        MPI_UINT64_T, toProc, DATA_TAG+rank, MPI_COMM_WORLD, &sendReq[sendReqCount++]);
            // printf("rank %d posted send to %d with size %d tag %d\n", rank, toProc, sendCnt[indexInArr], DATA_TAG+rank);
        }
    }
    int recvIndex = -1;
    while (countDown > 0) {
        // printf("rank %d recvReqCount %d countDown %d\n", rank, recvReqCount, countDown);
        MPI_Waitany(recvReqCount, recvReq, &recvIndex, &status);
        int count = -1;
        // printf ("rank %d MPI_Waitany returned %d.\n", rank, recvIndex);
        size_t indexInArr = (size_t) (status.MPI_SOURCE - procReceivesFrom[0]);
        MPI_Get_count(&status, MPI_UINT64_T, &count);
        #ifdef DEBUG
        assert(count >= 0);
        #endif
        recvCnt[indexInArr] = count;
        // printf("rank %d received actual job! from: %d size %d tag %d\n", rank, status.MPI_SOURCE, recvCnt[indexInArr], status.MPI_TAG);
        // Do the thing
        for (int iter1 = 0; iter1 < recvCnt[indexInArr] ; iter1+=2) {
            newp[p2precvData[indexInArr][iter1]-start_index] += p2precvData[indexInArr][iter1+1];

            // cascadingLocalUpdatesQueue[p2precvData[indexInArr][iter1]] += p2precvData[indexInArr][iter1+1];
            auto oldSize = valTable_cascadingLocalUpdatesQueue.size();
            valTable_cascadingLocalUpdatesQueue[p2precvData[indexInArr][iter1]] += p2precvData[indexInArr][iter1+1];
            if (oldSize != valTable_cascadingLocalUpdatesQueue.size())// not exists
                cascadingLocalUpdatesQueue.push(p2precvData[indexInArr][iter1]);
        }
        // printf("rank %d finished job from: %d size %d tag %d\n", rank, status.MPI_SOURCE, recvCnt[indexInArr], status.MPI_TAG);
        --countDown;
    }
    if (countOfCascade == 0)
        MPI_Allreduce(&totalSend, &globalSent, 1, MPI_LONG_LONG, MPI_MAX, MPI_COMM_WORLD);
    else
        MPI_Allreduce(&totalSend, &globalSent, 1, MPI_LONG_LONG, MPI_MAX, smallestKComm);

    // printf("rank %d totalSend[%d] %d globalSent in FUNC %d\n", rank, countOfCascade, totalSend, globalSent);

    // if (globalSent == 0) { // If there is no data to send at this point, break;
    //     return 0;
    // }

    #ifdef DEBUG
    if (globalSent < 0)
        std::cout << "Something is wrong! globalSent cannot be negative! rank " << rank << ", globalSent = " << globalSent << std::endl;
    #endif

    // for (int i=0; i < recvReqCount; ++i)
    //     MPI_Request_free(&recvReq[i]);
    for (int i=0; i < sendReqCount; ++i)
        MPI_Request_free(&sendReq[i]);

    delete[] recvReq;
    delete[] sendReq;

    // printf("rank %d END\n", rank);
    return globalSent;
}

// Returns globalSent

sizeType pointToPointComm (const int rank, const int start_index, const int countOfCascade,
                            MPI_Comm& smallestKComm,
                            std::vector<int>& procReceivesFrom, std::vector<int>& procSendsTo,
                            std::vector<int>& recvCnt, std::vector<int>& sendCnt,
                            std::vector<std::unique_ptr<uint64_t[]>>& p2precvData, std::vector<std::unique_ptr<uint64_t[]>>& p2psendData,
                            std::priority_queue < int, std::vector<int>, std::greater<int> >& cascadingLocalUpdatesQueue,
                            flat_hash_map <int, int>& valTable_cascadingLocalUpdatesQueue,
                            std::vector< flat_hash_map< int, int > >& valTable_cascadingOutOfProcUpd,
                            std::vector<int>& newp) {
    // First, Receive procedures
    MPI_Status status;
    sizeType totalSend = 0;
    sizeType globalSent = 0;
    int recvReqCount = 0, sendReqCount = 0;
    MPI_Request *recvReq = nullptr;
    MPI_Request *sendReq = nullptr;
    if (procReceivesFrom.size())
        recvReq = new MPI_Request[2*procReceivesFrom.size()];
    if (procSendsTo.size())
        sendReq = new MPI_Request[2*procSendsTo.size()];

    if (procReceivesFrom.size()) {
        for (int fromProc: procReceivesFrom) {
            // printf("rank %d receves from %d tag %d\n", rank, fromProc, COUNT_TAG+fromProc);
            size_t indexInArr = (size_t) (fromProc - procReceivesFrom[0]);
            MPI_Irecv(&recvCnt[indexInArr], 1, MPI_INT, fromProc, COUNT_TAG+fromProc, MPI_COMM_WORLD, &recvReq[recvReqCount++]);
            // MPI_Irecv(&recvCnt[fromProc - procReceivesFrom[0]], 1, MPI_INT, fromProc, COUNT_TAG+fromProc, MPI_COMM_WORLD, &recvReq[recvReqCount]);
        }
    }
    // printf ("rank %d posted MPI_Irecvs\n", rank);

    // Second, Send procedures
    for (int toProc: procSendsTo) {
        // printf("rank %d sends to %d tag %d\n", rank, toProc, COUNT_TAG+rank);
        size_t indexInArr = (size_t) (toProc - procSendsTo[0]);
        sendCnt[indexInArr] = 2 * valTable_cascadingOutOfProcUpd[toProc].size();
        if (sendCnt[indexInArr] > 0)
            p2psendData[indexInArr] = std::unique_ptr<uint64_t[]> (new uint64_t[sendCnt[indexInArr]]);
        totalSend += sendCnt[indexInArr];
        MPI_Isend(&sendCnt[indexInArr], 1, MPI_INT,
                    toProc, COUNT_TAG+rank, MPI_COMM_WORLD, &sendReq[sendReqCount++]);
        // int MPI_Isend(const void *buf, int count, MPI_Datatype datatype, int dest, int tag, MPI_Comm comm, MPI_Request *request)
    }
    // printf ("rank %d posted MPI_Isends\n", rank);
    for (int toProc: procSendsTo) {
        size_t itSend = 0;
        size_t indexInArr = (size_t) (toProc - procSendsTo[0]);
        if (sendCnt[indexInArr] > 0) {
            for(auto& mm: valTable_cascadingOutOfProcUpd[toProc]) {
                p2psendData[indexInArr][itSend++] = mm.first; // global node id
                p2psendData[indexInArr][itSend++] = mm.second; // value
            }
            MPI_Isend(&p2psendData[indexInArr][0], sendCnt[indexInArr],
                        MPI_INT, toProc, DATA_TAG+rank, MPI_COMM_WORLD, &sendReq[sendReqCount++]);
        }
    }
    size_t countDown = procReceivesFrom.size();
    int recvIndex = -1;

    while (countDown > 0) {
        // printf("rank %d recvReqCount %d\n", rank, recvReqCount);
        MPI_Waitany(recvReqCount, recvReq, &recvIndex, &status);
        // printf ("rank %d MPI_Waitany returned %d.\n",rank, recvIndex);
        // int MPI_Waitany(int count, MPI_Request array_of_requests[], int *index, MPI_Status *status)
        if ( (size_t) recvIndex < procReceivesFrom.size() ) { // Received is the size of update data
            // The request is from

            // printf("rank %d req %d, status-from:%d\n", rank, recvIndex, status.MPI_SOURCE);
            size_t indexInArr = (size_t) (status.MPI_SOURCE - procReceivesFrom[0]);
            // printf("rank %d src %d indexInArr %d\n", rank, status.MPI_SOURCE, indexInArr);
            if (recvCnt[indexInArr] > 0) {
                MPI_Irecv(&p2precvData[indexInArr][0], recvCnt[indexInArr], MPI_INT, status.MPI_SOURCE, DATA_TAG+status.MPI_SOURCE, MPI_COMM_WORLD, &recvReq[recvReqCount++]);
                // int MPI_Irecv(void *buf, int count, MPI_Datatype datatype, int source, int tag, MPI_Comm comm, MPI_Request *request)
            }
            else{
                --countDown;
            }
        }
        else { // The received is actual p update data
            // MPI_Request_get_status(recvReq[recvIndex], &flag, &status);
            // printf("rank %d received actual job! from: %d size %d tag %d\n", rank, status.MPI_SOURCE, recvCnt[status.MPI_SOURCE], status.MPI_TAG);
            --countDown;
            // int sourceHere = status.MPI_SOURCE;
            size_t indexInArr = (size_t) (status.MPI_SOURCE - procReceivesFrom[0]);
            // Do the thing
            for (auto iter1 = 0; iter1 < recvCnt[indexInArr] ; iter1+=2) {
                newp[p2precvData[indexInArr][iter1]-start_index] += p2precvData[indexInArr][iter1+1];

                // cascadingLocalUpdatesQueue[p2precvData[indexInArr][iter1]] += p2precvData[indexInArr][iter1+1];
                auto oldSize = valTable_cascadingLocalUpdatesQueue.size();
                valTable_cascadingLocalUpdatesQueue[p2precvData[indexInArr][iter1]] += p2precvData[indexInArr][iter1+1];
                if (oldSize != valTable_cascadingLocalUpdatesQueue.size())// not exists
                    cascadingLocalUpdatesQueue.push(p2precvData[indexInArr][iter1]);
            }
            // printf("rank %d finished job from: %d size %d tag %d\n", rank, status.MPI_SOURCE, recvCnt[status.MPI_SOURCE], status.MPI_TAG);

        }
    }

    if (countOfCascade == 0)
        MPI_Allreduce(&totalSend, &globalSent, 1, MPI_LONG_LONG, MPI_MAX, MPI_COMM_WORLD);
    else
        MPI_Allreduce(&totalSend, &globalSent, 1, MPI_LONG_LONG, MPI_MAX, smallestKComm);

    // for (int i=0; i < recvReqCount; ++i)
    //     MPI_Request_free(&recvReq[i]);
    // for (int i=0; i < sendReqCount; ++i)
    //     MPI_Request_free(&sendReq[i]);

    delete[] recvReq;
    delete[] sendReq;

    return globalSent;
}
sizeType allToAllComm (const int nProc, const int rank,
                            std::vector<int>& recvCnt, std::vector<int>& sendCnt,
                            std::vector<int>& recvDispl, std::vector<int>& sendDispl,
                            std::vector<int>& recvData, std::vector<int>& sendData,
                            std::vector< flat_hash_map< int, int > >& valTable_cascadingOutOfProcUpd
                            #ifdef PRINT
                            ,const int countOfCascade
                            #endif
                            ) {
    sizeType globalSent = 0;
    sendDispl[0] = 0;
    recvDispl[0] = 0;
    for (int iter1 = 0; iter1 < nProc; ++iter1) {
        sendCnt[iter1] = 2 * valTable_cascadingOutOfProcUpd[iter1].size();
        if (iter1 < DEF_lastProc)
            sendDispl[iter1+1] = sendDispl[iter1] + sendCnt[iter1];
    }
    sizeType totalSend = sendDispl[DEF_lastProc]+sendCnt[DEF_lastProc];

    MPI_Allreduce(&totalSend, &globalSent, 1, MPI_LONG_LONG, MPI_MAX, MPI_COMM_WORLD);
    // int MPI_Allreduce(const void *sendbuf, void *recvbuf, int count,
    //           MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)

    #ifdef PRINT
    std::cout << "rank " << rank << " totalSend " << totalSend << " globalSent " << globalSent << std::endl;
    #endif

    if (globalSent == 0) { // If there is no data to send at this point, break;
        return 0;
    }

    #ifdef DEBUG
    if (globalSent < 0) {
        std::cout << "Something is wrong! globalSent cannot be negative! rank " << rank << ", globalSent = " << globalSent << std::endl;
    }
    #endif

    #ifdef PRINT
    {
        std::cout << "cas " << countOfCascade << " rank " << rank << " sendCnt" << sendCnt << std::endl;
        std::cout << "cas " << countOfCascade << " rank " << rank << " sendDispl" << sendDispl << std::endl;
    }
    #endif

    // First, communicate the counts for each proc from each proc
    MPI_Alltoall(sendCnt.data(), 1, MPI_INT, recvCnt.data(), 1, MPI_INT, MPI_COMM_WORLD);
    // int MPI_Alltoall(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
    //                      void *recvbuf, int recvcount, MPI_Datatype recvtype,
    //                      MPI_Comm comm)

    for (int iter1 = 0; iter1 < DEF_lastProc; ++iter1) {
        recvDispl[iter1+1] = recvDispl[iter1] + recvCnt[iter1];
    }

    #ifdef PRINT
    {
        std::cout << "cas " << countOfCascade << " rank " << rank << " recvCnt" << recvCnt << std::endl;
        std::cout << "cas " << countOfCascade << " rank " << rank << " recvDispl" << recvDispl << std::endl;
    }
    #endif

    int totalRecv = recvDispl[DEF_lastProc]+recvCnt[DEF_lastProc];

    #ifdef PRINT
    {
        std::cout << "cas " << countOfCascade << " rank " << rank << " totalSend" << totalSend << std::endl;
        std::cout << "cas " << countOfCascade << " rank " << rank << " totalRecv" << totalRecv << std::endl;
    }
    #endif

    sendData.resize(totalSend);
    recvData.resize(totalRecv);

    for (int itSend = 0, iter2 = rank + 1; iter2 < nProc; ++iter2) {
        for(auto& mm: valTable_cascadingOutOfProcUpd[iter2]) {
            sendData[itSend] = mm.first; // global node id
            sendData[itSend+1] = mm.second; // value
            itSend += 2;
        }
    }

    #ifdef PRINT
    printf("ph1.6 (cas %d) rank %d sendData: ", countOfCascade, rank); for (int &x: sendData)printf("%2d ", x);printf("\n");
    #endif

    MPI_Alltoallv(sendData.data(), sendCnt.data(), sendDispl.data(), MPI_INT,
                    recvData.data(), recvCnt.data(), recvDispl.data(), MPI_INT, MPI_COMM_WORLD);
    // int MPI_Alltoallv(const void *sendbuf, const int *sendcounts,
    //                   const int *sdispls, MPI_Datatype sendtype, void *recvbuf,
    //                   const int *recvcounts, const int *rdispls, MPI_Datatype recvtype,
    //                   MPI_Comm comm)

    #ifdef PRINT
    {
        printf("Ph 1.5 cascade %d rank %d totalRecv %d\n", countOfCascade, rank, totalRecv);
        printf("Ph 1.5 cascade %d rank %d recvData: ", countOfCascade, rank); for (int &x: recvData)printf("%2d ", x);printf("\n");
    }
    #endif
    return globalSent;
}

sizeType allToAllSplitComm (const int nProc, const int rank,
                            std::vector<int>& recvCnt, std::vector<int>& sendCnt,
                            std::vector<int>& recvDispl, std::vector<int>& sendDispl,
                            std::vector<int>& recvData, std::vector<int>& sendData,
                            std::vector< flat_hash_map< int, int > >& valTable_cascadingOutOfProcUpd,
                            const int smallNProc, const int smallRank, MPI_Comm& smallestKComm
                            #ifdef PRINT
                            ,const int countOfCascade
                            #endif
) {
    sizeType globalSent = 0;
    const int gap = rank - smallRank;
    sendDispl.resize(smallNProc);
    sendCnt.resize(smallNProc);
    recvDispl.resize(smallNProc);
    recvCnt.resize(smallNProc);
    sendDispl[0] = 0;
    recvDispl[0] = 0;
    #ifdef DEBUG
    for (int iter1 = 0; iter1 < nProc; ++iter1) {
        if (valTable_cascadingOutOfProcUpd[iter1].size() > 0)
            if (((iter1 - rank) > (smallNProc - 1 - smallRank)) || (iter1 < rank)) {
                printf("rank %d smRank %d -- goal %d -- smallNProc %d How come I send to someone that far?\n", rank, smallRank, iter1, smallNProc);
                MPI_Abort(MPI_COMM_WORLD, 1);
                exit(EXIT_FAILURE);
            }
    }
    #endif

    for (int iter1 = 0; iter1 < smallNProc; ++iter1) {
        sendCnt[iter1] = 2 * valTable_cascadingOutOfProcUpd[iter1+gap].size();
        if (iter1 < smallNProc - 1)
            sendDispl[iter1+1] = sendDispl[iter1] + sendCnt[iter1];
    }
    sizeType totalSend = sendDispl[smallNProc-1]+sendCnt[smallNProc-1];

    MPI_Allreduce(&totalSend, &globalSent, 1, MPI_LONG_LONG, MPI_MAX, smallestKComm);
    // int MPI_Allreduce(const void *sendbuf, void *recvbuf, int count,
    //           MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)

    #ifdef PRINT
    std::cout << "rank " << rank << " (smRank " << smallRank << ") totalSend: " << totalSend << " globalSent " << globalSent << std::endl;
    #endif

    if (globalSent == 0) { // If there is no data to send at this point, break;
        return 0;
    }

    #ifdef DEBUG
    if (globalSent < 0) {
        std::cout << "Something is wrong! globalSent cannot be negative! rank " << rank << ", globalSent = " << globalSent << std::endl;
    }
    #endif

    #ifdef PRINT
    {
        std::cout << "cas " << countOfCascade << " rank " << rank << " (smRank " << smallRank << ") sendCnt: " << sendCnt << std::endl;
        std::cout << "cas " << countOfCascade << " rank " << rank << " (smRank " << smallRank << ") sendDispl: " << sendDispl << std::endl;
    }
    #endif

    // First, communicate the counts for each proc from each proc
    MPI_Alltoall(sendCnt.data(), 1, MPI_INT, recvCnt.data(), 1, MPI_INT, smallestKComm);
    // int MPI_Alltoall(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
    //                      void *recvbuf, int recvcount, MPI_Datatype recvtype,
    //                      MPI_Comm comm)
    for (int iter1 = 0; iter1 < smallNProc-1; ++iter1) {
        recvDispl[iter1+1] = recvDispl[iter1] + recvCnt[iter1];
    }

    #ifdef PRINT
    {
        std::cout << "cas " << countOfCascade << " rank " << rank << " (smRank " << smallRank << ") recvCnt: " << recvCnt << std::endl;
        std::cout << "cas " << countOfCascade << " rank " << rank << " (smRank " << smallRank << ") recvDispl: " << recvDispl << std::endl;
    }
    #endif

    int totalRecv = recvDispl[smallNProc-1]+recvCnt[smallNProc-1];

    #ifdef PRINT
    {
        std::cout << "cas " << countOfCascade << " rank " << rank << " (smRank " << smallRank << ") totalSend" << totalSend << std::endl;
        std::cout << "cas " << countOfCascade << " rank " << rank << " (smRank " << smallRank << ") totalRecv" << totalRecv << std::endl;
    }
    #endif


    sendData.resize(totalSend);
    recvData.resize(totalRecv);
    size_t itSend = 0;
    for (int iter2 = rank + 1; iter2 < nProc; ++iter2) {
        if (valTable_cascadingOutOfProcUpd[iter2].size() > 0 && (iter2 - rank) > (smallNProc - 1 - smallRank)) {
            printf("rank %d smRank %d -- goal %d -- smallNProc %d How come I send to someone that far?\n", rank, smallRank, iter2, smallNProc);
            MPI_Abort(MPI_COMM_WORLD, 1);
            exit(EXIT_FAILURE);
        }
        for(auto& mm: valTable_cascadingOutOfProcUpd[iter2]) {
            sendData[itSend] = mm.first; // global node id
            sendData[itSend+1] = mm.second; // value
            itSend += 2;
        }
        if (itSend > sendData.size())
            printf("rank %d smRank %d filed more than I should\n", rank, smallRank);
    }

    #ifdef PRINT
    printf("ph1.6 (cas %d) rank %d (smRank %d) sendData: ", countOfCascade, rank, smallRank); for (int &x: sendData)printf("%2d ", x);printf("\n");
    #endif


    MPI_Alltoallv(sendData.data(), sendCnt.data(), sendDispl.data(), MPI_INT,
                    recvData.data(), recvCnt.data(), recvDispl.data(), MPI_INT, smallestKComm);
    // int MPI_Alltoallv(const void *sendbuf, const int *sendcounts,
    //                   const int *sdispls, MPI_Datatype sendtype, void *recvbuf,
    //                   const int *recvcounts, const int *rdispls, MPI_Datatype recvtype,
    //                   MPI_Comm comm)

    // printf("rank %d (smRank %d) recvData: ", rank, smallRank); for (int &x: recvData)printf("%2d ", x);printf("\n");

    #ifdef PRINT
    {
        printf("Ph 1.5 cascade %d rank %d (smRank %d) totalRecv %d\n", countOfCascade, rank, smallRank, totalRecv);
        printf("Ph 1.5 cascade %d rank %d (smRank %d) recvData: ", countOfCascade, rank, smallRank); for (int &x: recvData)printf("%2d ", x);printf("\n");
    }
    #endif
    return globalSent;
}
