#ifndef _MPI_COMM_H
#define _MPI_COMM_H

#include <mpi.h>
#include <algorithm>
#include <vector>
#include <list>
#include <numeric>
#include <random>
#include <iostream>
#include <cstdio>
#include <cstdint>
#include <utility>
#include <queue>

#include "phmap.h"
#include "utils.hpp"
#include "io.hpp"
#include "graph.hpp"
using phmap::flat_hash_map;


#define COUNT_TAG 10000
#define DATA_TAG 20000

// Returns globalSent
sizeType pointToPointComm_SingleComm (const int rank, const uint64_t start_index, const int countOfCascade, const uint64_t countOfSmallestKLocal,
                            MPI_Comm& smallestKComm,
                            std::vector<int>& procReceivesFrom, std::vector<int>& procSendsTo,
                            std::vector<int>& recvCnt, std::vector<int>& sendCnt,
                            std::vector<std::unique_ptr<uint64_t[]>>& p2precvData, std::vector<std::unique_ptr<uint64_t[]>>& p2psendData,
                            std::priority_queue < uint64_t, std::vector<uint64_t>, std::greater<uint64_t> >& cascadingLocalUpdatesQueue,
                            flat_hash_map <uint64_t, int>& valTable_cascadingLocalUpdatesQueue,
                            std::vector< flat_hash_map< uint64_t, int > >& valTable_cascadingOutOfProcUpd,
                            std::vector<int>& newp);

sizeType pointToPointComm (const int rank, const int start_index, const int countOfCascade,
                            MPI_Comm& smallestKComm,
                            std::vector<int>& procReceivesFrom, std::vector<int>& procSendsTo,
                            std::vector<int>& recvCnt, std::vector<int>& sendCnt,
                            std::vector<std::unique_ptr<uint64_t[]>>& p2precvData, std::vector<std::unique_ptr<uint64_t[]>>& p2psendData,
                            std::priority_queue < int, std::vector<int>, std::greater<int> >& cascadingLocalUpdatesQueue,
                            flat_hash_map <int, int>& valTable_cascadingLocalUpdatesQueue,
                            std::vector< flat_hash_map< int, int > >& valTable_cascadingOutOfProcUpd,
                            std::vector<int>& newp);

sizeType allToAllComm (const int nProc, const int rank,
                            std::vector<int>& recvCnt, std::vector<int>& sendCnt,
                            std::vector<int>& recvDispl, std::vector<int>& sendDispl,
                            std::vector<int>& recvData, std::vector<int>& sendData,
                            std::vector< flat_hash_map< int, int > >& valTable_cascadingOutOfProcUpd
                            #ifdef PRINT
                            ,const int countOfCascade
                            #endif
                            );

sizeType allToAllSplitComm (const int nProc, const int rank,
                            std::vector<int>& recvCnt, std::vector<int>& sendCnt,
                            std::vector<int>& recvDispl, std::vector<int>& sendDispl,
                            std::vector<int>& recvData, std::vector<int>& sendData,
                            std::vector< flat_hash_map< int, int > >& valTable_cascadingOutOfProcUpd,
                            const int smallNProc, const int smallRank, MPI_Comm& smallestKComm
                            #ifdef PRINT
                            ,const int countOfCascade
                            #endif
                            );
#endif
