#include <mpi.h>
#include <algorithm>
#include <vector>
#include <list>
#include <numeric>
#include <random>
#include <iostream>
#include <cstdio>
#include <cstdint>
#include <utility>
#include <queue>

#include "phmap.h"
#include "mpi_generator.h"
#include "mpi_comm.h"
#include "utils.hpp"
#include "io.hpp"
#include "graph.hpp"

using namespace std;
using phmap::flat_hash_set;
using phmap::flat_hash_map;

template <class Generator, class T, class TT>
void Phase2and3Optimized(vector<uint64_t>& S, vector<int>& p, const uint64_t start_index,
                        Generator& gen, graph<T,TT>& G, std::unique_ptr <uint64_t[]>& RsizeUsed) {
    auto &xadj = G.xadj;
    auto &adj = G.adj;

    flat_hash_set<uint64_t> Set;
    Set.reserve(S.size()+1);
    vector<uint64_t> R;
    R.reserve(S.size());
    uint64_t _i = 0, _j = 0, i = 0, j = 0, r = 0;
    while(!S.empty()) {

        i = S.back();
        S.pop_back();

        _i = i - start_index;

        #ifdef PRINT
        std::cout << "i " << i << " -> p[i] = " << p[_i] << std::endl;
        #endif
        // r = uniform_int_distribution<int>(max(0, (int)S.size() - p[_i]), (int)S.size())(gen);
        r = pickSubsetSize( ((int)S.size() - p[_i]), (int)S.size(), gen);
        R.resize(r);
        subsetsample(0, (int)S.size(), R.begin(), R.end(), Set, gen);
        for(auto it: R) {
            j = S[it];
            _j = j - start_index;
            if(i < j) {
                #ifdef PRINT
                std::cout << "added edge " << i << " -> " << j << std::endl;
                #endif
                adj[xadj[_i] + RsizeUsed[_i]] = j;
                ++RsizeUsed[_i];
            }
            else {
                #ifdef PRINT
                std::cout << "added edge " << j << " -> " << i << std::endl;
                #endif
                adj[ xadj[_j] + RsizeUsed[_j] ] = i;
                ++RsizeUsed[_j];
            }
            p[_j]++;
        }
        for(int it = 0; it < (int)S.size(); it++) {
            j = S[it];
            _j = j - start_index;
            p[ _j ]--;
            if(p[ _j ] == 0) {

                #ifdef PRINT
                std::cout << "j " << j << " -> p[j] = " << p[_j] << std::endl;
                #endif

                swap(S[it], S.back());
                // r = S.back();
                S.pop_back();
                for(auto k: S) {
                    if(j < k) {
                        #ifdef PRINT
                        std::cout << "added edge " << j << " -> " << k << std::endl;
                        #endif
                        adj[ xadj[_j] + RsizeUsed[_j] ] = k;
                        ++RsizeUsed[_j];
                    }
                    else {
                        #ifdef PRINT
                        std::cout << "added edge " << k << " -> " << j << std::endl;
                        #endif
                        adj[ xadj[k - start_index] + RsizeUsed[k - start_index]] = j;
                        ++RsizeUsed[k - start_index];
                    }
                }
                it--;
            }
        }
    }

}

graph <uint64_t, sizeType> generateGraphWithShellSequence(std::vector<int>& begin,
                    const uint64_t size, const uint64_t n, vector<uint64_t>& hist_starti,
                    vector<uint64_t>& all_start_indices, const int comm_mode, const int randomSeed,
                    timeType* times) {
    double commstart;
    double phstart;

    int _rank, _nProc;
    MPI_Comm_size(MPI_COMM_WORLD, &_nProc);
    MPI_Comm_rank(MPI_COMM_WORLD, &_rank);
    const int rank = _rank;
    const int nProc = _nProc;

    // Variables for COMM_ALL2ALL_SPLIT
    MPI_Comm smallestKComm;
    int smallNProc = -1;
    int smallRank = -1;
    if (comm_mode > COMM_ALL2ALL_BELOW) {
        MPI_Comm_split(MPI_COMM_WORLD, begin[0], rank, &smallestKComm);
        // int MPI_Comm_split(MPI_Comm comm, int color, int key, MPI_Comm* newcomm);
        MPI_Comm_rank(smallestKComm, &smallRank);
        MPI_Comm_size(smallestKComm, &smallNProc);
    }
    // Variables for COMM_ALL2ALL_SPLIT end

    bool lastCoreAlreadyDone = 0;

    random_device rd;  //Will be used to obtain a seed for the random number engine
    int _genSeed = rd();
    mt19937_64 gen(_genSeed); //Standard mersenne_twister_engine seeded with rd()
    if (randomSeed > 0){
        gen.seed(randomSeed+rank);
        // printf("seed at %d ---> %d\n", rank, randomSeed+rank);
    }
    // else {
    //     printf("seed at %d ---> %d\n", rank, _genSeed);
    // }

    #ifdef PRINT
    std::cout << "rank " << rank << " hist_starti: " << hist_starti << std::endl;
    #endif

    auto end = begin.end();

    const int highestK = hist_starti.size() - 2;
    const int localHighestK = begin[size-1];

    uint64_t _i = -1, _j = -1;

    graph <uint64_t, sizeType> G;

    G.N = size;
    G.xadj = std::unique_ptr<sizeType[]>(new sizeType[size + 1]);

    auto &N = G.N;
    auto &xadj = G.xadj;

    std::unique_ptr<uint64_t[]> RsizeUsed = std::unique_ptr<uint64_t[]>(new uint64_t[N]);

    // ssum --> (max) number of edges a processor may generate
    sizeType ssum = 0;
    auto arriter = begin.begin();
    xadj[0] = 0;
    _i = 1;
    while (arriter != end){
        ssum += *arriter;
        xadj[_i++] = ssum;
        ++arriter;
    }

    G.adj = std::unique_ptr<uint64_t[]>(new uint64_t[ssum]);
    auto &adj = G.adj;
    // printf("rank %d ssum %d\n",rank, ssum);

    uint64_t start_index = all_start_indices[rank];

    // printf("rank %d all_start_indices: ", rank);for(int& j:all_start_indices) printf("%d ", j);printf("\n");

    vector< priority_queue < uint64_t, std::vector<uint64_t>, std::greater<uint64_t> > > cascadingOutOfProcUpd(nProc);
    vector< flat_hash_map< uint64_t, int > > valTable_cascadingOutOfProcUpd(nProc);
    vector <int> p(size, 0);
    vector <int> newp(size); // for a copy of p

    int lastProcHoldingHighestOfCurPlusOne = upper_bound(&all_start_indices[rank+1], &all_start_indices[nProc], hist_starti[localHighestK+1]-1) - &(all_start_indices[0]);
    // printf("r%d lastProcHoldingHighestOfCurPlusOne %d\n", rank, lastProcHoldingHighestOfCurPlusOne);
    double t_start = MPI_Wtime();

    uint64_t lastIndexToLook = n-highestK-1;
    if (rank == DEF_lastProc) {
        for (uint64_t i = lastIndexToLook; i < n; ++i) {
            RsizeUsed[i - start_index] = 0;
        }
    }

    vector <int> procReceivesFrom;  // Variable to hold proc ids current processor may receive data from.
    vector <int> procSendsTo;       // Variable to hold proc ids current processor may send data to.
                                    // Used for COMM_ONE2ONE.
    int inisize_sendsto;
    int inisize_recvfrom;
    if (comm_mode >= COMM_POINT_TO_POINT_ABOVE) {
        int smallestLocalK = begin[0]; // first shell val
        uint64_t startOfSmallestLocalK = hist_starti[smallestLocalK];
        int locationOfStart = upper_bound(all_start_indices.begin(), all_start_indices.end(), startOfSmallestLocalK) - all_start_indices.begin() - 1;
        procReceivesFrom.resize(rank - locationOfStart);
        iota(procReceivesFrom.begin(),procReceivesFrom.end(), locationOfStart);
        inisize_recvfrom = procReceivesFrom.size();
        // procReceivesFrom hold the range [locationOfStart, rank)

        // int largestLocalK = begin[size-1]; // last shell val
        uint64_t endOfLargestLocalK = hist_starti[localHighestK+1] - 1; // last index before the first index of next shell val
        int locationOfEnd = upper_bound(all_start_indices.begin(), all_start_indices.end(), endOfLargestLocalK) - all_start_indices.begin() - 1;
        // printf("rank %d smallK %d startOfSmallestLocalK %d locationOfStart %d\nrank %d largeK %d endOfLargestLocalK %d locationOfEnd %d\n", rank, smallestLocalK, startOfSmallestLocalK, locationOfStart, rank, largestLocalK, endOfLargestLocalK, locationOfEnd);
        procSendsTo.resize(locationOfEnd - rank);
        iota(procSendsTo.begin(),procSendsTo.end(), rank+1);
        inisize_sendsto = procSendsTo.size();
        // printf("rank %d procReceivesFrom: ", rank);for (int &x: procReceivesFrom)printf("%2d ", x);printf("\n");
        // printf("rank %d procSendsTo: ", rank);for (int &x: procSendsTo)printf("%2d ", x);printf("\n");
    }


    // ============================
    // ========= PHASE 1 ==========
    // ============================


    if ( lastIndexToLook != 0 ) {

        flat_hash_set<uint64_t> Set;
        Set.reserve(localHighestK);
        // While we are not at last Kmax+1 elements
        for (_i=0; (_i + start_index) < lastIndexToLook && (_i < size); ++_i) {
            #ifdef DEBUG
            assert(begin[_i] >= 0); assert(p[_i]>=0);
            #endif
            subsetsample( _i+start_index+1, n, &adj[xadj[_i]], &adj[xadj[_i+1]], Set, gen );
        }
        times->subset += MPI_Wtime() - t_start;

        int currentK = -1;
        uint64_t lastIndexWithSameK = -1;

        for (_i=0; (_i + start_index) < lastIndexToLook && (_i < size); ++_i) {
            // int i = start_index + _i;
            if (begin[_i] != currentK) { // we moved to next K value in the local elements of this processor
                currentK = begin[_i];
                lastIndexWithSameK = hist_starti[currentK+1] - 1; // higher core's start point - 1 is the end point of ours.
            }

            #ifdef DEBUG
            assert(p[_i] >= 0);
            #endif

            RsizeUsed[_i] = pickSubsetSize(begin[_i] - p[_i], begin[_i], gen);

            #ifdef PRINT
            std::cout << "rank " << rank << " i " <<  _i+start_index << " ss " << begin[_i] << " RsizeUsed[_i] (" << begin[_i] - p[_i] << "<=)" << RsizeUsed[_i] << "(<=" << begin[_i] << ")" << std::endl;
            #endif

            for (uint64_t k = 0; k < RsizeUsed[_i]; ++k) {
                uint64_t j = adj[xadj[_i]+k];

                #ifdef PRINT
                std::cout << "rank " << rank << " R+:{" << j << "}" << std::endl;
                #endif

                #ifdef DEBUG
                assert(j > _i+start_index);
                #endif

                if (j <= lastIndexWithSameK) {
                    _j = j - start_index;
                    if (_j < size) {
                        ++p[ _j ];
                    }
                    else {
                        // int _proc_to_send2 = upper_bound(all_start_indices.begin(), all_start_indices.end(), j) - all_start_indices.begin() - 1;
                        int _proc_to_send = upper_bound(&all_start_indices[rank+2], &all_start_indices[lastProcHoldingHighestOfCurPlusOne], j) - &(all_start_indices[1]);
                        auto oldSize = valTable_cascadingOutOfProcUpd.size();
                        ++valTable_cascadingOutOfProcUpd[_proc_to_send][j];
                        if (oldSize != valTable_cascadingOutOfProcUpd.size()) { // not exists
                            cascadingOutOfProcUpd[_proc_to_send].push(j);
                        }
                    }
                }
            }
            // printf("rank %d p: ", rank);for (int &x: p)printf("%2d ", x);printf("\n");
        }

        times->phase1 += MPI_Wtime() - t_start;

        // CHECK IF THE LAST CORE FITS IN SINGLE PROCESSOR.
        // IF SO, COMPLETE THAT AS WELL BEFORE MOVING ON TO COMMUNICATION
        // Phase 2
        if (rank == DEF_lastProc) {
            // If first index of last core is >= start index of last proc,
            // then we can go ahead and do this.
            if (hist_starti[highestK] >= all_start_indices[rank]) {
                std::cout <<"Last core fits in single proc. Running it before fix procedure." << std::endl;
                phstart = MPI_Wtime();
                vector <uint64_t> S (highestK+1);
                iota(S.begin(),S.end(), n-highestK-1);
                for (uint64_t i=0, j=S.size()-1; i < j; ++i)
                    if (p[ S[i] - start_index] == 0)
                        swap(S[i--], S[j--]);

                Phase2and3Optimized(S, p, start_index, gen, G, RsizeUsed);
                times->phase2 += MPI_Wtime() - phstart;
                lastCoreAlreadyDone = 1;
            }
        }

        // if (comm_mode == NO_COMM_GUARANTEED) break or jump to end
        if (comm_mode != NO_COMM_GUARANTEED) {

            vector <int> sendCnt;
            vector <int> recvCnt;

            // // For A2A comm
            // vector <int> sendDispl;
            // vector <int> recvDispl;

            // vector <uint64_t> sendData;
            // vector <uint64_t> recvData;

            // For P2P comm
            vector<std::unique_ptr<uint64_t[]>> p2psendData;
            vector<std::unique_ptr<uint64_t[]>> p2precvData;

            uint64_t countOfSmallestKLocal = min(hist_starti[begin[0]+1] - all_start_indices[rank], size);
            if (hist_starti[begin[0]] == all_start_indices[rank]) // cannot receive anything since it is the first index for the K.
                countOfSmallestKLocal = 0;
            // printf ("r %d smK %d countOfSmallestKLocal %d\n", rank, begin[0], countOfSmallestKLocal);

            // if (comm_mode < COMM_POINT_TO_POINT_ABOVE) {
            //     sendCnt.resize(nProc);
            //     recvCnt.resize(nProc);
            //     sendDispl.resize(nProc);
            //     recvDispl.resize(nProc);
            //     recvData.resize(nProc*2*countOfSmallestKLocal);
            // }
            // else
            if (comm_mode >= COMM_POINT_TO_POINT_ABOVE) {
                sendCnt.resize(procSendsTo.size());
                recvCnt.resize(procReceivesFrom.size());
                p2psendData.resize(procSendsTo.size());
                p2precvData.resize(procReceivesFrom.size());
                for (size_t i=0; i < procReceivesFrom.size(); ++i) {
                    p2precvData[i] = std::unique_ptr<uint64_t[]> (new uint64_t[2*countOfSmallestKLocal]);
                }
            }

            int countOfCascade = 0;
            double fixStart = MPI_Wtime();

            sizeType globalSent = 0; // 1 to allow first entrance to loop. instead of do-while

            while (1) {
                // loop exit condition is when globalSent is zero. Checked in the middle.

                priority_queue < uint64_t, std::vector<uint64_t>, std::greater<uint64_t> > cascadingLocalUpdatesQueue;
                flat_hash_map <uint64_t, int> valTable_cascadingLocalUpdatesQueue;

                // vector <int> newp(p); // start with a copy of p; update values as newp
                copy(p.begin(), p.end(), newp.begin());

                // ============================
                // ======== PHASE 1.1 =========
                // ============================

                commstart = MPI_Wtime();
                if (comm_mode > COMM_POINT_TO_POINT_ABOVE) {
                    if (comm_mode == COMM_ONE2ONE_BETTER) {
                        globalSent = pointToPointComm_SingleComm(rank, start_index, countOfCascade, countOfSmallestKLocal,
                                smallestKComm,
                                procReceivesFrom, procSendsTo,
                                recvCnt, sendCnt,
                                p2precvData, p2psendData,
                                cascadingLocalUpdatesQueue,
                                valTable_cascadingLocalUpdatesQueue,
                                valTable_cascadingOutOfProcUpd,
                                newp);
                    }
                    // else if (comm_mode == COMM_ONE2ONE) {
                    //     globalSent = pointToPointComm(rank, start_index, countOfCascade,
                    //             smallestKComm,
                    //             procReceivesFrom, procSendsTo,
                    //             recvCnt, sendCnt,
                    //             p2precvData, p2psendData,
                    //             cascadingLocalUpdatesQueue,
                    //             valTable_cascadingLocalUpdatesQueue,
                    //             valTable_cascadingOutOfProcUpd,
                    //             newp);
                    // }
                    else {
                        std::cout << "Invalid comm_mode! (" << comm_mode << ")" << std::endl;
                        MPI_Abort(MPI_COMM_WORLD, 1);
                        exit(EXIT_FAILURE);
                    }
                }
                // else if ( (comm_mode < COMM_ALL2ALL_BELOW) ||
                //             // ( (comm_mode == COMM_ALL2ALL_SPLIT || comm_mode == COMM_ALL2ALL_SPLIT_BETTER) && countOfCascade == 0)
                //             (countOfCascade == 0)
                //         ) {
                //     globalSent = allToAllComm(nProc, rank,
                //             recvCnt, sendCnt,
                //             recvDispl, sendDispl,
                //             recvData, sendData,
                //             valTable_cascadingOutOfProcUpd
                //             #ifdef PRINT
                //             ,countOfCascade
                //             #endif
                //             );
                // }
                // else if ( (comm_mode > COMM_ALL2ALL_BELOW && comm_mode < COMM_POINT_TO_POINT_ABOVE) && countOfCascade > 0) {
                //     // After first round (countOfCascade > 0), we can just divide to groups and do Alltoallv in their respective subgroups.
                //     // There cannot be any processes in the same group for a single k. All Alltoallv's can be done in parallel for each k group.
                //     // TODO: Try to find a (cheap) way to exclude first process of each communicator, so that idle processors don't wair for nothing.
                //     if (comm_mode == COMM_ALL2ALL_SPLIT) {
                //         globalSent = allToAllSplitComm(nProc, rank,
                //                 recvCnt, sendCnt,
                //                 recvDispl, sendDispl,
                //                 recvData, sendData,
                //                 valTable_cascadingOutOfProcUpd,
                //                 smallNProc, smallRank, smallestKComm
                //                 #ifdef PRINT
                //                 ,countOfCascade
                //                 #endif
                //                 );
                //     }
                //     else {
                //         printf("Invalid comm_mode! (%d)\n", comm_mode);
                //         MPI_Abort(MPI_COMM_WORLD, 1);
                //         exit(EXIT_FAILURE);
                //     }
                // }
                else {
                    std::cout << "Invalid comm_mode! (" << comm_mode << ")" << std::endl;
                    MPI_Abort(MPI_COMM_WORLD, 1);
                    exit(EXIT_FAILURE);
                }
                #ifdef PRINT
                std::cout << "rank " << rank << " cas " << countOfCascade << " globalSent " << globalSent << std::endl;
                #endif
                if (globalSent == 0) { // If there is no data to send at this point, break;
                    break;
                }

                #ifdef DEBUG
                assert (globalSent > 0);
                #endif
                if (comm_mode >= COMM_POINT_TO_POINT_ABOVE) {
                    // Check procSendsTo first. we change procReceivesFrom.size() in the other if
                    if (procSendsTo.size() > 0) {
                        if (begin[0] != localHighestK) // smallest k == largest,
                                        // means it is receiving and sending regarding the same k
                                        // otherwise, receiving for k_small and sending for k_large
                            procSendsTo.resize(0); // then, don't need to send to anyone in the next...
                        if (procReceivesFrom.size() == 0) // was not receiving from anyone
                            procSendsTo.resize(0); // then, don't need to send to anyone in the next...
                    }
                    if (procReceivesFrom.size() > 0) // was receiving from someone
                        procReceivesFrom.erase(procReceivesFrom.begin()); // remove first receive
                }

                // ============================
                // ======== PHASE 1.5 =========
                // ============================

                // This phase does the local updates according to received data
                // in Phase 1.1 and re-populates the remoteUpdate maps (hashmap, pq)

                // fixStuff() begin

                // Traverse the recvData array to get all local info to pq and set.
                // COMM_ONE2ONE does it right after receiving
                // if (comm_mode < COMM_POINT_TO_POINT_ABOVE) {
                //     for (size_t iter1 = 0; iter1 < recvData.size() ; iter1+=2) {
                //         newp[recvData[iter1]-start_index] += recvData[iter1+1];

                //         auto oldSize = valTable_cascadingLocalUpdatesQueue.size();
                //         valTable_cascadingLocalUpdatesQueue[recvData[iter1]] += recvData[iter1+1];
                //         if (oldSize != valTable_cascadingLocalUpdatesQueue.size())// not exists
                //             cascadingLocalUpdatesQueue.push(recvData[iter1]);
                //     }
                // }
                times->communications += MPI_Wtime() - commstart;

                // cascadingOutOfProcUpd: variable to hold the new cascading updates that need to be sent to other processors.
                cascadingOutOfProcUpd = vector< priority_queue < uint64_t, std::vector<uint64_t>, std::greater<uint64_t> > > (nProc);
                valTable_cascadingOutOfProcUpd = vector< flat_hash_map<uint64_t, int> > (nProc);

                ++countOfCascade;

                #ifdef PRINT
                for(auto& mm: valTable_cascadingLocalUpdatesQueue)
                    std::cout << "Ph 1.5 cascade " << countOfCascade << " rank " << rank << " priority_queue [" << mm.first << "] = " << mm.second << std::endl;
                #endif

                int currentK = -1;
                uint64_t lastIndexWithSameK = -1;

                while (!cascadingLocalUpdatesQueue.empty()) {

                    uint64_t i = cascadingLocalUpdatesQueue.top();
                    cascadingLocalUpdatesQueue.pop();
                    _i = i - start_index;

                    #ifdef DEBUG
                    assert(_i < size);
                    #endif

                    if (newp[_i] < 0) {
                        #ifdef PRINT
                        std::cout << "Cascade " << countOfCascade << " check this out: node _" << _i << " has newp[_i] = " << newp[_i] << std::endl;
                        #endif
                        newp[_i] = 0;
                    }

                    if (i >= lastIndexToLook) {
                        p[_i] = newp[_i];
                        #ifdef PRINT
                        std::cout << "Not an error: We defer the updates on these (i = " << i << " > " << (n-highestK-1) << " (n-highestK-1) )until Phase 2 and 3" << std::endl;
                        #endif
                        continue;
                    }
                    if (newp[_i] == p[_i]) {
                        #ifdef PRINT
                        std::cout << "No change in total removed and added edges from earlier indices." << std::endl;
                        #endif
                        continue;
                    }

                    if (p[_i] >= begin[_i] && newp[_i] >= begin[_i]) {
                        #ifdef PRINT
                        std::cout << "Lucky!!!! i = " << i << " p[i] (" << p[_i] << ") was already >= k (" << begin[_i] << ") and newp[i] (" << newp[_i] << ")" << std::endl;
                        #endif
                        p[_i] = newp[_i];
                        // equivalently, if the range was 0,kmax and new is also 0,kmax;
                        // we can skip this update.
                        continue;
                    }
                    // Update p with newp for the later iterations of the loop.
                    p[_i] = newp[_i];

                    uint64_t repickedSize = pickSubsetSize(begin[_i] - newp[_i], begin[_i], gen);

                    #ifdef PRINT
                    std::cout << "Ph1.5 rank " << rank << " i " << i << " repickedSize " << repickedSize << " RsizeUsed " << RsizeUsed[_i] << std::endl;
                    #endif

                    if (repickedSize == RsizeUsed[_i]) {

                        #ifdef PRINT
                        std::cout << "rank " << rank << " Lucky me! No updates for this one (" << i << ") (" << RsizeUsed[_i] << "->"  << repickedSize << ")!" << std::endl;
                        #endif
                        continue;
                    }

                    if (begin[_i] != currentK) { // we moved to next K value in the local elements of this processor
                        currentK = begin[_i];
                        lastIndexWithSameK = hist_starti[currentK+1] - 1; // higher core's start point - 1 is the end point of ours.
                    }

                    auto [smaller, larger] = minmax(repickedSize, RsizeUsed[_i]);
                    // auto [smaller, larger] = minmax(repickedSize, RsizeUsed[_i]);
                    int change = (RsizeUsed[_i] < repickedSize ? 1 : -1);

                    // for the nodes in randomsubset of _i, iterate over new added ones and check p equivalency.
                    // update anything new, and add stuff to cascadingLocalUpdatesQueue for ...
                    for (size_t k = smaller; k < larger; ++k) {
                        uint64_t j = adj[xadj[_i]+k];

                        #ifdef PRINT
                        if (change > 0) std::cout << "rank " << rank << " R[" << i << "] :PLUS {" << j << "}" << std::endl;
                        else std::cout << "rank " << rank << " R[" << i << "] :MINUS {" << j << "}" << std::endl;;
                        #endif

                        if (j <= lastIndexWithSameK) {

                            _j = j - start_index;

                            if (_j < size) {
                                newp[ _j ] += change;
                                auto oldSize = valTable_cascadingLocalUpdatesQueue.size();
                                valTable_cascadingLocalUpdatesQueue[j] += change;
                                if (oldSize != valTable_cascadingLocalUpdatesQueue.size())// not exists
                                    cascadingLocalUpdatesQueue.push(j);
                            }
                            else {
                                // find the processor that contains j
                                // int _proc_to_send2 = upper_bound(all_start_indices.begin(), all_start_indices.end(), j) - all_start_indices.begin() - 1;
                                int _proc_to_send = upper_bound(&all_start_indices[rank+2], &all_start_indices[lastProcHoldingHighestOfCurPlusOne], j) - &(all_start_indices[1]);

                                auto oldSize = valTable_cascadingOutOfProcUpd.size();
                                valTable_cascadingOutOfProcUpd[_proc_to_send][j] += change;
                                if (oldSize != valTable_cascadingOutOfProcUpd.size()) { // not exists
                                    cascadingOutOfProcUpd[_proc_to_send].push(j);
                                }
                            }
                        }
                    }

                    RsizeUsed[_i] = repickedSize;

                } // while (!cascadingLocalUpdatesQueue.empty()) end

                // printf("rank %d queue is empty now.\n", rank);

                // fixStuff() end

            } // while (1) end
            times->fix += MPI_Wtime() - fixStart;
            times->phase1 += times->fix;
            #ifdef PRINT
            std::cout << "rank " << rank << " No more updates from other processors" << std::endl;
            #endif

            times->countOfCascade = countOfCascade;

        }
        // else {
        //     printf("No Comm\n");
        // }
    }
    if (rank == DEF_lastProc) {
        if ( !lastCoreAlreadyDone ) {
            std::cout << "Last core does not fit in single proc. Running it after fix procedure." << std::endl;
            phstart = MPI_Wtime();
            vector <uint64_t> S (highestK+1);
            iota(S.begin(),S.end(), n-highestK-1);
            for (uint64_t i=0, j=S.size()-1; i < j; ++i)
                if (p[ S[i] - start_index] == 0)
                    swap(S[i--], S[j--]);

            Phase2and3Optimized(S, p, start_index, gen, G, RsizeUsed);
            times->phase2 += MPI_Wtime() - phstart;
        }
    }

    // generateFullCSR();
    phstart = MPI_Wtime();
    G.fixCSR(&RsizeUsed[0]); // RsizeUsed is of length N, not N+1. Need to send &start and &start+size
    times->fixCSR += MPI_Wtime() - phstart;
    // generateFullCSR(); end

    // Timer stopped before print out
    times->generator = MPI_Wtime() - t_start;
    if (comm_mode > COMM_ALL2ALL_BELOW)
        MPI_Comm_free(&smallestKComm);

    return G;
}
