#ifndef _MPIUTILS_HPP
#define _MPIUTILS_HPP
#include <mpi.h>

#include <algorithm>
#include <vector>
#include <list>
#include <numeric>
#include <random>
#include <iostream>
#include <cstdio>
#include <cstdint>
#include <utility>
#include <queue>

#include "phmap.h"
#include "utils.hpp"
#include "io.hpp"
#include "graph.hpp"

template <class vIDType, class eCountType>
int distributedVerifier (graph <vIDType, eCountType>& fullG, std::vector<vIDType>& hist_ps) {
    // TO BE IMPLMENTED
    return 0;
}

template <class vIDType, class eCountType>
graph <vIDType, eCountType> fullCSR (graph <vIDType, eCountType>& G, std::vector<vIDType>& all_start_indices, const int rank, const int nProc) {

    graph <vIDType, eCountType> fullG;
    fullG.N = G.N;

    fullG.xadj = std::unique_ptr<eCountType[]>(new eCountType[fullG.N + 1]);
    auto& newXadj = fullG.xadj;
    newXadj[0] = 0;

    vIDType myStart = all_start_indices[rank];
    std::vector< phmap::flat_hash_map< vIDType, std::vector<uint64_t> > > toSend(nProc);

    for (vIDType _i = 0; _i < G.N; ++_i) {
        for (eCountType j = G.xadj[_i]; j < G.xadj[_i+1]; ++j) {
            int _proc_to_send = std::upper_bound(all_start_indices.begin(), all_start_indices.end(), G.adj[j]) - all_start_indices.begin() - 1;
            if (_proc_to_send != rank)
                toSend[_proc_to_send][G.adj[j]].push_back(_i + myStart);
        }
    }

    // TODO: Replace this with something scalable
    std::vector <int> recvCnt(nProc);
    std::vector <int> sendCnt(nProc);
    std::vector <int> recvDispl(nProc);
    std::vector <int> sendDispl(nProc);
    std::unique_ptr<uint64_t[]> recvData;
    std::unique_ptr<uint64_t[]> sendData;

    // func: sum of number of edges to sent to each process
    //                          node id + edge count + edges = 2 + b.second.size()
    auto accumulator = [](eCountType a, const auto& b) { return a + 2 + b.second.size(); };

    sendDispl[0] = 0;
    for (int iter1 = 0; iter1 < nProc; ++iter1) {
        sendCnt[iter1] = std::accumulate(toSend[iter1].begin(), toSend[iter1].end(), uint64_t(), accumulator);
        if (iter1 < DEF_lastProc)
            sendDispl[iter1+1] = sendDispl[iter1] + sendCnt[iter1];
    }
    // std::cout << "R " << rank << " sendDispl: " << sendDispl << std::endl;
    // std::cout << "R " << rank << " sendCnt: " << sendCnt << std::endl;

    MPI_Alltoall(sendCnt.data(), 1, MPI_INT, recvCnt.data(), 1, MPI_INT, MPI_COMM_WORLD);

    recvDispl[0] = 0;
    for (int iter1 = 0; iter1 < DEF_lastProc; ++iter1) {
        recvDispl[iter1+1] = recvDispl[iter1] + recvCnt[iter1];
    }

    // std::cout << "my " << rank << " recvDispl[rank]: " << recvDispl << std::endl;
    // std::cout << "my " << rank << " recvCnt[rank]: " << recvCnt << std::endl;

    eCountType totalEdge = 2 * G.xadj[G.N]; // local edge count * 2
    // std::cout << "rank " << rank << " totalEdge*2 " << totalEdge << std::endl;
    if (rank != DEF_lastProc) {
        // totalEdge = G.xadj[G.N];
        // for (vIDType _i = 0; _i < G.N; ++_i) {
        //     for (eCountType j = G.xadj[_i]; j < G.xadj[_i+1]; ++j) {
        //         if (G.adj[j] < myStart + G.N)
        //             ++totalEdge;
        //     }
        // }
        // std::cout << "rank " << rank << " totalEdge_counted_f " << totalEdge << std::endl;
        totalEdge = G.xadj[G.N];
        for (vIDType _i = 0; _i < G.N; ++_i)
            totalEdge += std::count_if ( &G.adj[G.xadj[_i]], &G.adj[G.xadj[_i+1]],
                                        [&] (vIDType n) { return n < myStart + G.N; } );

        // std::cout << "rank " << rank << " totalEdge_counted_s " << totalEdge << std::endl;
    }

    phmap::flat_hash_map< uint64_t, std::vector<vIDType> > allRecv;
    // priority_queue < vIDType, std::vector<vIDType>, std::greater<vIDType> > incoming;

    vIDType totalRecv = recvDispl.back() + recvCnt.back();
    {
        recvData = std::unique_ptr<uint64_t[]> (new uint64_t[totalRecv]); // +1 for empty send
        {
            vIDType totalSend = sendDispl.back() + sendCnt.back();
            sendData = std::unique_ptr<uint64_t[]> (new uint64_t[totalSend]); // +1 for empty send
            eCountType itSend = 0;
            for (int iter1 = 0; iter1 < nProc; ++iter1) {
                for(auto &m: toSend[iter1]) {
                    sendData[itSend++] = m.first;
                    sendData[itSend++] = m.second.size();
                    // sendData.insert(...)
                    for (vIDType n: m.second)
                        sendData[itSend++] = n;
                }
            }
            // std::cout << "R " << rank << " sendData " << sendData << std::endl;
            MPI_Alltoallv(&sendData[0], sendCnt.data(), sendDispl.data(), MPI_UINT64_T,
                            &recvData[0], recvCnt.data(), recvDispl.data(), MPI_UINT64_T, MPI_COMM_WORLD);
        }
        // std::cout << "R " << rank << " recvData " << recvData << std::endl;

        for (eCountType i=0; i < totalRecv; ) {
            // if (allRecv.find(recvData[i] - myStart) == allRecv.end()) {
                // incoming.push(recvData[i] - myStart);
            // }
            auto& v = allRecv[recvData[i] - myStart]; // if it does not exist, this will generate one
            ++i;
            v.reserve(recvData[i]); // count
            totalEdge += recvData[i]; //
            ++i;
            v.insert(v.end(), &recvData[i], &recvData[i+recvData[i-1]]);
            i += recvData[i-1];
        }
    }

    // std::cout << "R " << rank << " totalEdge " << totalEdge << std::endl;
    fullG.adj = std::unique_ptr<vIDType[]>(new vIDType[totalEdge]);
    auto& newAdj = fullG.adj;

    for (vIDType i=0; i < G.N; ++i) {
        if (G.xadj[i] < G.xadj[i+1]) {
            // if (allRecv.find(i) == allRecv.end()) {
            //     incoming.push(i);
            // }
        }
        auto& v = allRecv[i];
        v.reserve(G.xadj[i+1] - G.xadj[i] + v.size());
        v.insert(v.end(), &G.adj[G.xadj[i]], &G.adj[G.xadj[i+1]]);
        for (eCountType j = G.xadj[i]; j < G.xadj[i+1]; ++j) {
            allRecv[G.adj[j] - myStart].push_back(i + myStart);
            // std::cout << "R " << rank  << " i " << i + myStart << " v " << v;
            // std::cout << " newv " << v <<std::endl;
        }
        std::sort(v.begin(), v.end());
    }
    for (vIDType i=0; i < G.N; ++i) {
        newXadj[i+1] = newXadj[i] + allRecv[i].size();
    }
    // if (rank == DEF_lastProc)
    for (vIDType cur = 0; cur < fullG.N; ++ cur) {
        // vIDType cur = incoming.top();
        auto& v = allRecv[cur];
        // Assuming uint64_t
        memcpy(&newAdj[newXadj[cur]], v.data(), v.size() * sizeof(vIDType));

    }
    // if (rank == DEF_lastProc) {
    //     std::cout << "xadj: ";
    //     for (vIDType i=0; i <= G.N; ++i) {
    //         std::cout << " " << newXadj[i];
    //     }
    //     std::cout << std::endl << "adj: ";
    //     for (vIDType i=0; i < newXadj[G.N]; ++i) {
    //         std::cout << " " << newAdj[i];
    //     }
    //     std::cout << std::endl;
    // }

    return fullG;
}

#endif
