#ifndef SEQUENTIAL_HPP
#define SEQUENTIAL_HPP

#include <sstream>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <numeric>
#include <iterator>
#include <vector>
#include <cassert>
#include <array>
#include <random>
#include <string>
#include <random>
#include <memory>
#include <mutex> //for timer

#include "phmap.h"
using phmap::flat_hash_set;

#include "utils.hpp"
#include "graph.hpp"

template<class vIDType, class eCountType, class Generator, class Container, class Container1>
inline void phase2and3(Container &t, Container1 &f, graph<vIDType, eCountType> &G, flat_hash_set<vIDType> &Set, Generator &gen, vIDType K, vIDType offset = 0) {

	std::vector<vIDType> S(K + 1);
	std::iota(S.begin(), S.end(), G.N - K - 1);

	for(vIDType i = 0, j = S.size() - 1; i < j; i++)
		if(t[S[i]] == 0)
			std::swap(S[i--], S[j--]);

	std::vector<vIDType> R;
	R.reserve(S.size());

	while(!S.empty()) {
		auto i = S.back();
		S.pop_back();
		auto r = std::uniform_int_distribution<vIDType>(std::max(t[i], (vIDType)S.size()) - t[i], (vIDType)S.size())(gen);
		R.resize(r);
		subsetsample((vIDType)0, (vIDType)S.size(), R.begin(), R.end(), Set, gen, /* needShuffle */false);
		for(auto j: R) {
			j = S[j];
			if(i < j)
				G.adj[G.xadj[i] + f[i]++] = j + offset;
			else
				G.adj[G.xadj[j] + f[j]++] = i + offset;
			t[j]++;
		}
		for(vIDType j = 0; j < (vIDType)S.size(); j++) {
			t[S[j]]--;
			if(t[S[j]] == 0) {
				std::swap(S[j], S.back());
				auto x = S.back();
				S.pop_back();
				for(auto y: S)
					if(x < y)
						G.adj[G.xadj[x] + f[x]++] = y + offset;
					else
						G.adj[G.xadj[y] + f[y]++] = x + offset;
				j--;
			}
		}
	}
}

template<class vIDType, class eCountType = std::uint64_t>
inline auto sequentialSSGG(std::vector<vIDType> hist) {
	std::mutex iomutex;
	timer tim("sequentialSSGG", iomutex);

	std::random_device rnd;
	std::mt19937 gen(rnd());

	auto shellseq = histToShellSeq(hist);
	graph<vIDType, eCountType> G(shellSeqToOutdegs(shellseq));

	auto &xadj = G.xadj;
	auto &adj = G.adj;
	auto N=G.N;

	std::vector<vIDType> f(N);
	std::vector<vIDType> t(N);
	{
		flat_hash_set<vIDType> Set;
		Set.reserve(shellseq.back());

		{
			timer tim("phase1: subsetselection & main algorithm", iomutex);
			for(vIDType i = 0; i < N - shellseq.back() - 1; i++) {
				f[i] = std::uniform_int_distribution<vIDType>(std::max(t[i], shellseq[i]) - t[i], shellseq[i])(gen);
				subsetsample(i + 1, N, adj.get() + xadj[i], adj.get() + xadj[i] + f[i], Set, gen, /* needShuffle */false);
				for(auto j = xadj[i]; j < xadj[i] + f[i]; j++)
					if(shellseq[i] == shellseq[adj[j]])
						t[adj[j]]++;
			}
			std::cerr << "times.phase1 : " << tim.time() << '\n';
		}
		{
			timer tim("phase2", iomutex);
			phase2and3(t, f, G, Set, gen, shellseq.back());
			std::cerr << "times.phase2 : " << tim.time() << '\n';
		}
	}

	{
		timer tim("CSR fixing phase", iomutex);
		G.fixCSR(f.begin());
        std::cerr << "times.fixCSR : " << tim.time() << '\n';
	}
	std::cerr << "times.all : " << tim.time() << '\n';

	return G;
}


#endif